﻿///////////////////////////////////////////////////////////////////////////////
//
//    (C) Copyright 2012 EaseFilter Technologies Inc.
//    All Rights Reserved
//
//    This software is part of a licensed software product and may
//    only be used or copied in accordance with the terms of that license.
//
///////////////////////////////////////////////////////////////////////////////

using System;
using System.Text;
using System.IO;
using System.Diagnostics;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Threading;
using Microsoft.Win32.SafeHandles;
using System.Globalization;

namespace CSharpDemo
{
    public partial class FilterDemo : Form
    {
        const uint FILE_FLAG_NO_BUFFERING = 0x20000000;
        const uint FILE_FLAG_OPEN_REPARSE_POINT = 0x00200000;
        const int AlignedBufferSize = 65536; //normally the sector size 512 bytes, 65536 is big enough for all disk's sector size.

        [DllImport("kernel32.dll")]
        private static extern bool WriteFile(IntPtr hFile, IntPtr lpBuffer, int NumberOfBytesToWrite, out int lpNumberOfBytesWritten, IntPtr lpOverlapped);

        [DllImport("kernel32.dll")]
        private static extern bool ReadFile(IntPtr hFile, IntPtr lpBuffer, int nNumberOfBytesToRead, out int lpNumberOfBytesRead, IntPtr lpOverlapped);

        /// <summary>
        /// The requirement for FILE_FLAG_NO_BUFFERING read and write.
        /// Aligned buffer is the buffer size must be an integer multiple of the volume sector size.
        /// Buffer addresses for read and write operations should be physical sector-aligned, 
        /// which means aligned on addresses in memory that are integer multiples of the volume's physical sector size. 
        /// Depending on the disk, this requirement may not be enforced.
        /// </summary>
        /// <param name="alignedBuffer">This the alignedBuffer address, we alway use the fix buffer size 65536</param>
        /// <returns>return the buffer which allocated and it needs to be released after it completed</returns>
        public static IntPtr AllocateAlignedBuffer(out IntPtr alignedBuffer)
        {
            IntPtr buffer = Marshal.AllocHGlobal(AlignedBufferSize * 2);
            
            alignedBuffer = buffer;

            for (int i = 0; i < AlignedBufferSize; i++)
            {
                alignedBuffer = new IntPtr(buffer.ToInt64() + i);

                if (alignedBuffer.ToInt64() % AlignedBufferSize == 0)
                {
                    break;
                }
                
            }

            return buffer;
        }

        public static byte[] ConvertHexStringToByteArray(string hexString)
        {
            if (hexString.Length % 2 != 0)
            {
                throw new ArgumentException(String.Format(CultureInfo.InvariantCulture, "The binary key cannot have an odd number of digits: {0}", hexString));
            }

            byte[] HexAsBytes = new byte[hexString.Length / 2];
            for (int index = 0; index < HexAsBytes.Length; index++)
            {
                string byteValue = hexString.Substring(index * 2, 2);
                HexAsBytes[index] = byte.Parse(byteValue, NumberStyles.HexNumber, CultureInfo.InvariantCulture);
            }

            return HexAsBytes;
        }

        public static string ConvertByteArrayToHexString(byte[] ba)
        {
            string retVal = string.Empty;

            for (int i = 0; i < ba.Length; i++)
            {
                retVal += string.Format("{0:x2}", ba[i]);
            }

            return retVal;
        }

        private bool GetEncryptionKeyAndIV(ref byte[] key, ref byte[] iv)
        {
            string keyStr = textBox_EncryptKey.Text;
            string ivStr = textBox_IV.Text;

            iv = null;

            key = ConvertHexStringToByteArray(keyStr);
            if (ivStr.Length > 0)
            {
                iv = ConvertHexStringToByteArray(ivStr);
            }

            if (key != null && key.Length != 16 && key.Length != 24 && key.Length != 32)
            {
                MessageBox.Show("The encryption key length must be 16,24 or 32 bytes", "Key Length", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            if (iv != null && iv.Length != 16)
            {
                MessageBox.Show("The initial vector length must be 16 bytes!", "IV Length", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            return true;
        }

        private void button_EncryptFile_Click(object sender, EventArgs e)
        {
            string fileName = textBox_EncryptFileName.Text;
         
            byte[] key = null;
            byte[] iv = null;
            uint ivLength = 0;

            if (!GetEncryptionKeyAndIV(ref key, ref iv))
            {
                return;
            }

            if (iv != null)
            {
                ivLength = (uint)iv.Length;
            }

            bool retVal = EaseFilter.AESEncryptFile(fileName, (uint)key.Length, key,ivLength,iv,checkBox_AddIVTag.Checked);
            if (!retVal)
            {
                string errorMessage = EaseFilter.GetLastErrorMessage();
                textBox_EncryptionInfo.Text = "Encrypt file " + fileName + " failed with error:" + errorMessage;
            }
            else
            {
                textBox_EncryptionInfo.Text = "Encrypt file " + fileName + " succeeded.";
            }
        }


        private void button_CopyEncryptedFile_Click(object sender, EventArgs e)
        {
            string fileName = textBox_EncryptFileName.Text;
            byte[] iv = null;
            string lastError = string.Empty;
            bool retVal = EaseFilter.GetIVTag(fileName,ref iv,out lastError);

            if (!retVal)
            {
               textBox_EncryptionInfo.Text = "Get IVTag failed for " + fileName + " return error:" + lastError;
               return;
            }

            if (iv == null || iv.Length == 0)
            {
                textBox_EncryptionInfo.Text = "There are no iv tag in the file " + fileName + ", we can't copy it.";
                return;
            }

            string newFileName =Path.GetDirectoryName(fileName) + "\\" +  Path.GetFileNameWithoutExtension(fileName) + ".copy" + Path.GetExtension(fileName);

            if (DialogResult.OK == Input.InputBox("Copy Encrypt File To", "Input destination file path:", ref newFileName))
            {
              
                IntPtr sourceFileHandle = IntPtr.Zero;
                IntPtr destFileHandle = IntPtr.Zero;

                //open the file without buffering, it make sure the memory cache is clear text.
                //open file with reparse point option, it can by pass filter driver, even the filter driver is not enabled,we can open it to read.

                bool ret = EaseFilter.CreateFileAPI(fileName, (uint)FileAccess.Read, (uint)FileShare.Read, (uint)FileMode.Open, FILE_FLAG_NO_BUFFERING | FILE_FLAG_OPEN_REPARSE_POINT, ref sourceFileHandle);
                if (!ret)
                {
                    lastError = EaseFilter.GetLastErrorMessage();
                    textBox_EncryptionInfo.Text ="Open file " +  fileName + " failed with error:" + lastError;
                    return;
                }

                SafeFileHandle shFile = new SafeFileHandle(sourceFileHandle, true);

                ret = EaseFilter.CreateFileAPI(newFileName, (uint)FileAccess.Write, (uint)FileShare.None, (uint)FileMode.CreateNew, FILE_FLAG_NO_BUFFERING | FILE_FLAG_OPEN_REPARSE_POINT, ref destFileHandle);
                if (!ret)
                {
                    shFile.Close();

                    lastError = EaseFilter.GetLastErrorMessage();
                    textBox_EncryptionInfo.Text = "Create destination file " + newFileName + " failed with error:" + lastError;
                    return;
                }

                SafeFileHandle destFile = new SafeFileHandle(destFileHandle, true);

                IntPtr alignedBuffer;
                IntPtr bufferPtr = AllocateAlignedBuffer(out alignedBuffer);

                do
                {
                    //since we open file with no buffering, we have to read and write with integer multiple sector sizes.

                    int readWriteReturnLength = 0;
                    
                    if( ReadFile(sourceFileHandle, alignedBuffer, AlignedBufferSize, out readWriteReturnLength, IntPtr.Zero))
                    {
                        if (readWriteReturnLength > 0)
                        {
                            if (!WriteFile(destFileHandle, alignedBuffer, AlignedBufferSize, out readWriteReturnLength, IntPtr.Zero))
                            {
                                MessageBox.Show("Write file failed with error code " + Marshal.GetLastWin32Error(), "Write File.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                                break;
                            }
                        }
                        else
                        {
                            break;
                        }
                    }
                    else
                    {
                        MessageBox.Show("Read file failed with error code " + Marshal.GetLastWin32Error(), "Read File.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        break;
                    }

                } while (true);

                FileInfo fileInfo = new FileInfo(fileName);

                //set the destination file size to the same as source file
                if (!EaseFilter.SetFileSize(destFileHandle, fileInfo.Length))
                {
                    lastError = EaseFilter.GetLastErrorMessage();
                    textBox_EncryptionInfo.Text = "Set file size for file " + newFileName + " failed with error:" + lastError;
                }

                shFile.Close();
                destFile.Close();
                Marshal.FreeHGlobal(bufferPtr);

                //add the ivTag to the destination file, then the filter driver knows this file was encrypted.
                if (!EaseFilter.AddIVTag(newFileName, (uint)iv.Length, iv))
                {
                    textBox_EncryptionInfo.Text = "Copy encrypted file " + fileName + " to " + newFileName + " failed with error " + EaseFilter.GetLastErrorMessage();
                }
                else
                {
                    textBox_EncryptionInfo.Text = "Copy encrypted file " + fileName + " to " + newFileName + " succeeded.";
                }
            }
        }    


        private void button_DecryptFile_Click(object sender, EventArgs e)
        {
            string fileName = textBox_EncryptFileName.Text;

            byte[] key = null;
            byte[] iv = null;
            uint ivLength = 0;

            if (!GetEncryptionKeyAndIV(ref key, ref iv))
            {
                return;
            }

            if (iv != null)
            {
                ivLength = (uint)iv.Length;
            }

            bool retVal = EaseFilter.AESDecryptFile(fileName, (uint)key.Length, key, ivLength, iv);

            if (!retVal)
            {
                string errorMessage = EaseFilter.GetLastErrorMessage();
                textBox_EncryptionInfo.Text = "Decrypt file " + fileName + " failed with error:" + errorMessage;
            }
            else
            {
                textBox_EncryptionInfo.Text = "Decrypt file " + fileName + " succeeded.";
            }
        }

        private void button_AddEncryptionFilterRule_Click(object sender, EventArgs e)
        {
            EaseFilter.AccessFlag accessFlag = EaseFilter.AccessFlag.FILE_ENCRYPTION_RULE;
            string encryptionFilterRule = textBox_EncryptionFilterMask.Text;

            byte[] encryptionKey = ConvertHexStringToByteArray(textBox_EncryptFilterRuleKey.Text);
            if (encryptionKey != null && encryptionKey.Length != 16 && encryptionKey.Length != 24 && encryptionKey.Length != 32)
            {
                MessageBox.Show("The encryption key length must be 16,24 or 32 bytes", "Key Length", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            configSetting.FilterType = EaseFilter.FilterType.FILE_SYSTEM_ENCRYPTION;

            if (!configSetting.AddFilterRule(accessFlag, encryptionFilterRule, "", encryptionKey))
            {
                System.Windows.Forms.MessageBox.Show("Filter rule already exist.");
            }
        }

        private void button_RemoveEncryptionFilterRule_Click(object sender, EventArgs e)
        {
            foreach (System.Windows.Forms.ListViewItem item in listView_EncryptFilterRule.SelectedItems)
            {
                configSetting.RemoveFilterRule(item);
            }

        }

        private void button_ApplyEncryptionSetting_Click(object sender, EventArgs e)
        {
            System.Collections.Generic.List<ConfigSetting.FilterRule> filterRuleList = configSetting.FilterList;
            foreach (ConfigSetting.FilterRule filterRule in filterRuleList)
            {
                byte[] encryptionKey = filterRule.EncryptionKey;
                EaseFilter.AddFilterRule((uint)filterRule.AccessFlag, filterRule.FilterMask, filterRule.ReparseMask, (uint)encryptionKey.Length, encryptionKey);
            }

            configSetting.RequestRegistration = 0;
        }

        private byte[] ReadFileRaw(string fileName, int readLength)
        {
            //if the file was encrypted, there are reparse point added to the file, without the filter driver enable,it can't be read normally.
            //To test an encrypted file content if it was encrypted, you need to open the file with reparse point option and no buffer in cache manager.          

            string lastError = string.Empty;
            IntPtr fileHandle = IntPtr.Zero;

            bool ret = EaseFilter.CreateFileAPI(fileName, (uint)FileAccess.Read, (uint)FileShare.Read, (uint)FileMode.Open, FILE_FLAG_NO_BUFFERING | FILE_FLAG_OPEN_REPARSE_POINT, ref fileHandle);
            if (!ret)
            {
                lastError = EaseFilter.GetLastErrorMessage();
                throw new Exception(lastError);
            }

            SafeFileHandle shFile = new SafeFileHandle(fileHandle, true);

            IntPtr alignedBuffer;
            IntPtr bufferPtr = AllocateAlignedBuffer(out alignedBuffer);

            int numberOfBytesRead = 0;

            if (!ReadFile(fileHandle, alignedBuffer, AlignedBufferSize, out numberOfBytesRead, IntPtr.Zero))
            {
                MessageBox.Show("Read file failed with error code " + Marshal.GetLastWin32Error(), "Read File.", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            shFile.Close();

            byte[] buffer = new Byte[numberOfBytesRead];

            Marshal.Copy(alignedBuffer, buffer, 0, numberOfBytesRead);

            return buffer;
        }

        private void button_ReadEncryptedFile_Click(object sender, EventArgs e)
        {
           string fileName = textBox_EncryptFileName.Text;
           
            byte[] buffer = ReadFileRaw(fileName,65536);

            string rawStr = string.Empty;

            if (checkBox_RawHex.Checked)
            {
                rawStr = ConvertByteArrayToHexString(buffer);
            }
            else
            {
                rawStr = Encoding.Default.GetString(buffer);
            }

            textBox_EncryptionInfo.Text = rawStr;

        }

        private void button_DeleteIVTag_Click(object sender, EventArgs e)
        {
            string fileName = textBox_EncryptFileName.Text;

            if (MessageBox.Show("If the IV tag was deleted,the filter driver can't process the encrypted file, are you sure to delete the IV tag?", "Delete IV tag", MessageBoxButtons.YesNo, MessageBoxIcon.Warning) != DialogResult.OK)
            {
                return;
            }

            if (!EaseFilter.DeleteIVTag(fileName))
            {
                textBox_EncryptionInfo.Text = "Delete iv tag for file " + fileName + " failed with error " + EaseFilter.GetLastErrorMessage();
            }
            else
            {
                textBox_EncryptionInfo.Text = "Delete iv tag for file " + fileName + " succeeded.";
            }
        }


        private void button_EncryptionUnitTest_Click(object sender, EventArgs e)
        {
            try
            {
                //start the encryption unit test here......

                //stop filter service first if it was started.
                EaseFilter.StopFilter();

                //uninstall the filter driver here if it was already installed 
                EaseFilter.UnInstallDriver();


                //Create test folder c:\EncryptFolder
                string testFolder = "c:\\encryptFolder";
                if (!Directory.Exists(testFolder))
                {
                    Directory.CreateDirectory(testFolder);
                }

                //create test file here
                string testFile = Path.Combine(testFolder, "test0.txt");
                if (File.Exists(testFile))
                {
                    File.Delete(testFile);
                }

                //Add clear text data to the file.
                string iv = "f0f1f2f3f4f5f6f7f8f9fafbfcfdfeff";// Initialization vector with hex string
                string clearText = "6bc1bee22e409f96e93d7e117393172a";// clear text with hex string

                //256 bit test key
                string key256 = "603deb1015ca71be2b73aef0857d77811f352c073b6108d72d9810a30914dff4";// encrytpion key with hex string
                string cipherText256 = "601ec313775789a5b7a7f504bbf3d228";// cipher text with hex string

                byte[] testData = ConvertHexStringToByteArray(clearText);

                File.WriteAllBytes(testFile, testData);

                textBox_EncryptionUnitTestInfo.Text = "Test file " + testFile + " created." + Environment.NewLine +  Environment.NewLine;

                //Now we will encrypt this file with the API,
                byte[] encryptionKey = ConvertHexStringToByteArray(key256);
                byte[] ivTag = ConvertHexStringToByteArray(iv);

                bool ret = EaseFilter.AESEncryptFile(testFile, (uint)encryptionKey.Length, encryptionKey,(uint)ivTag.Length,ivTag,true);
                if (!ret)
                {
                    textBox_EncryptionUnitTestInfo.Text += "Encrypt file " + testFile + " failed with error " + EaseFilter.GetLastErrorMessage() + Environment.NewLine +  Environment.NewLine;
                    return;
                }

                textBox_EncryptionUnitTestInfo.Text += testFile + " was encrypted with 256 bit key." + Environment.NewLine + Environment.NewLine;

                string lastError = string.Empty;

                ret = EaseFilter.GetIVTag(testFile,ref ivTag, out lastError);
                if (!ret)
                {
                    textBox_EncryptionUnitTestInfo.Text += "Get iv tag for encrypted file " + testFile + " failed:" + lastError  + Environment.NewLine + Environment.NewLine;
                    return;
                }

                string ivString = ConvertByteArrayToHexString(ivTag);
                if (string.Compare(ivString, iv) == 0)
                {
                    textBox_EncryptionUnitTestInfo.Text += "Verified encrypted file " + testFile + " iv tag passed." + Environment.NewLine + Environment.NewLine;
                }
                else
                {
                    textBox_EncryptionUnitTestInfo.Text += "Verified encrypted file " + testFile + " iv tag failed, the correct iv is " + iv + " return iv:" + ivString + Environment.NewLine + Environment.NewLine;
                    return;
                }                
                
                //now we need to verify the raw encrypted data.
                byte[] rawData = ReadFileRaw(testFile, testData.Length);
                string encryptedText = ConvertByteArrayToHexString(rawData);

                if (string.Compare(cipherText256, encryptedText) == 0)
                {
                    textBox_EncryptionUnitTestInfo.Text += "Verified " + testFile + " 256 bit key encrypted data passed." + Environment.NewLine + Environment.NewLine;
                }
                else
                {
                    textBox_EncryptionUnitTestInfo.Text += "Verified " + testFile + " 256 bit key encrypted data failed,the correct cipher text is " + cipherText256 + "\r\nreturn cipher text is:" + encryptedText + Environment.NewLine + Environment.NewLine;
                    return;
                }


                //Now install the encryption filter driver
                if (!EaseFilter.InstallDriver())
                {
                    textBox_EncryptionUnitTestInfo.Text += "Install Driver Failed with error:" + EaseFilter.GetLastErrorMessage() + Environment.NewLine + Environment.NewLine;
                    return;
                }
                else
                {
                    textBox_EncryptionUnitTestInfo.Text += "Install Driver succeeded." + Environment.NewLine + Environment.NewLine;
                }


                //Add encryption filter rule
                //clear the filter rule
                configSetting.FilterList.Clear();
                configSetting.InitListView();
                configSetting.FilterType = EaseFilter.FilterType.FILE_SYSTEM_ENCRYPTION;

                EaseFilter.AccessFlag accessFlag = EaseFilter.AccessFlag.FILE_ENCRYPTION_RULE;
                string encryptFolder = testFolder + "\\*";
                if (!configSetting.AddFilterRule(accessFlag, encryptFolder, "", encryptionKey))
                {
                    textBox_EncryptionUnitTestInfo.Text += "Add encryption filter rule Failed. " + Environment.NewLine + Environment.NewLine;
                    return;
                }

                //don't register any callback I/O for this test,only for filter type control filter is enabled.
                configSetting.RequestRegistration = 0;

                 //start the filter driver service.
                button_Start_Click(null, null);

                if (!EaseFilter.IsFilterStarted)
                {
                    textBox_EncryptionUnitTestInfo.Text += "Start Filter Driver Service Failed " + Environment.NewLine +  Environment.NewLine;
                    return;

                }
                
                textBox_EncryptionUnitTestInfo.Text += "Filter driver service started." + Environment.NewLine + Environment.NewLine;

                //Now we will decrypt the file with the API
                ret = EaseFilter.AESDecryptFile(testFile, (uint)encryptionKey.Length, encryptionKey, 0, null);
                if (!ret)
                {
                    textBox_EncryptionUnitTestInfo.Text += "Decrypt file " + testFile + " failed with error " + EaseFilter.GetLastErrorMessage() + Environment.NewLine + Environment.NewLine;
                    return;
                }

                byte[] decryptedRawData = ReadFileRaw(testFile, testData.Length);
                string decryptedRawDataStr = ConvertByteArrayToHexString(decryptedRawData);

                if (string.Compare(clearText, decryptedRawDataStr) == 0)
                {
                    textBox_EncryptionUnitTestInfo.Text += "Verified decrypted file " + testFile + " passed."  + Environment.NewLine + Environment.NewLine;
                }
                else
                {
                    textBox_EncryptionUnitTestInfo.Text += "Verified decrypted file " + testFile + " failed. The correct clear test is " + clearText + "\r\nreturn text:" + decryptedRawDataStr + EaseFilter.GetLastErrorMessage() + Environment.NewLine + Environment.NewLine;
                    return;
                }

                textBox_EncryptionUnitTestInfo.Text += "Now you can star the encryption test, when you add new files to the folder " + testFolder + ",the new created files will be encrypted by filter driver." + Environment.NewLine +  Environment.NewLine;
                textBox_EncryptionUnitTestInfo.Text += "When you stop the filter driver,the encrypted files can't be read." + Environment.NewLine + Environment.NewLine;
            }
            catch (Exception ex)
            {
                textBox_EncryptionUnitTestInfo.Text += "Encryption unit test failed with error " + ex.Message + Environment.NewLine +  Environment.NewLine;
            }

        }


        private void button_StopEncryptionUnitTest_Click(object sender, EventArgs e)
        {
            //stop filter service first if it was started.
            EaseFilter.StopFilter();

            //uninstall the filter driver here if it was already installed 
            if (!EaseFilter.UnInstallDriver())
            {
                textBox_EncryptionUnitTestInfo.Text += "UnInstall Driver Failed with error:" + EaseFilter.GetLastErrorMessage() + Environment.NewLine + Environment.NewLine;
            }

            textBox_EncryptionUnitTestInfo.Text += "UnInstall Driver succeeded, unit test stopped." + EaseFilter.GetLastErrorMessage() + Environment.NewLine + Environment.NewLine;

        }    
   
    }
}
